@@ -2,7 +2,7 @@
 CheckedDict
 -----------
 
-A dictionary based on a default prototype. A CheckedDict can only define
+A dictionary based on a default prototype. A :class:`CheckedDict` can only define
 ``key:value`` pairs which are already present in the default. It is possible to
 define a docstring for each key and different restrictions for the values
 regarding possible values, ranges and type. A CheckedDict is useful for
@@ -12,15 +12,17 @@ configuration settings.
 ConfigDict
 ----------
 
-Based on CheckedDict, a ConfigDict is a persistent, unique dictionary. It is
+Based on :class:`CheckedDict`, a :class:`ConfigDict` is a persistent, unique dictionary. It is
 saved under the config folder determined by the OS and it is updated with each
 modification. It is useful for implementing configuration of a module / library
 / app, where there is a default/initial state and the user needs to be able to
 configure global settings which must be persisted between sessions (similar to
 the settings in an application)
 
-Example::
+Example
+~~~~~~~
 
+.. code::
 
     from configdict import ConfigDict
 
@@ -31,8 +33,9 @@ Example::
     config.load()
 
 
-Alternatively, a ConfigDict can be created all at once::
+Alternatively, a :class:`ConfigDict` can be created all at once
 
+.. code::
 
     config = ConfigDict("myapp",
         default = {
@@ -57,7 +60,7 @@ This will create the dictionary and load any persisted version. Any saved
 modifications will override the default values. Whenever the user changes any
 value (via ``config[key] = newvalue``) the dictionary will be saved.
 
-In all other respects a ConfigDict behaves like a normal dictionary.
+In all other respects a :class:`ConfigDict` behaves like a normal dictionary.
 
 """
 from __future__ import annotations
@@ -282,7 +285,8 @@ def _openInEditor(cfg):
 class CheckedDict(dict):
     """
     A dictionary which checks that the keys and values are valid
-    according to a default dict and a validator.
+    according to a default dict and a validator. In a :class:`CheckedDict`,
+    only keys are allowed which are already present in the default given.
 
     Args:
         default: a dict will all default values. A config can accept only
@@ -304,7 +308,27 @@ class CheckedDict(dict):
         docs: a dict containing help lines for keys defined in default
         callback: function ``(key, value) -> None``. This function is called **after**
             the modification has been done.
-        precallback: function ``(key, value) -> None``
+        precallback: function ``(key, value) -> newvalue``. If given, a precallback intercepts
+            any change and can modify the value or return None to prevent the modification
+
+    Example
+    =======
+
+    .. code::
+
+        from configdict import *
+        default = {
+            'color': '#FF0000',
+            'size': 10,
+            'name': ''
+        }
+
+        validator = {
+            'size::range': (6, 30),
+            'color': lambda d, value: iscolor(value)
+        }
+
+        checked = CheckedDict(default, validator=validator)
     """
     def __init__(self,
                  default: Dict[str, Any] = None,
@@ -357,8 +381,10 @@ class CheckedDict(dict):
         default config item by item (see example). After adding all new keys it is
         necessary to call ``.load()``
 
-        Example::
+        Example
+        =======
 
+        .. code::
 
             cfg = ConfigDict("foo", load=False)
             # We define a default step by step
@@ -439,7 +465,7 @@ class CheckedDict(dict):
 
     def getValidateFunc(self, key:str) -> Opt[validatefunc_t]:
         """
-        Returns a function to validate a value for `key`, if there
+        Returns a function to validate a value for ``key``, if there
         is one. A validate function has the form ``(config, value) -> bool``
 
         Args:
@@ -455,8 +481,7 @@ class CheckedDict(dict):
 
     def getChoices(self, key: str) -> Opt[list]:
         """
-        Return a seq. of possible values for key `k`
-        or `None`
+        Return a seq. of possible values for key ``k`` or ``None``
         """
         if key not in self._allowedkeys:
             raise KeyError(f"{key} is not a valid key")
@@ -482,7 +507,10 @@ class CheckedDict(dict):
 
         Returns errormsg. If value is of correct type, errormsg is None
 
-        Example::
+        Example
+        =======
+
+        .. code::
 
             error = config.checkType(key, value)
             if error:
@@ -526,11 +554,13 @@ class CheckedDict(dict):
         Returns the expected type for key, as a type which can be passed
         to isinstance
 
-        **NB**: all numbers are reduced to type float, all strings are of type str,
+        .. note::
+
+            All numbers are reduced to type float, all strings are of type str,
             otherwise the type of the default value, which can be a collection
             like a list or a dict
 
-        See Also: `checkValue`
+        See Also: :meth:`checkValue`
         """
         if self._validator is not None:
             definedtype = self._validator.get(key+"::type")
@@ -596,12 +626,12 @@ class CheckedDict(dict):
         """
         return value if value is not None else self.get(key, default)
 
-    def asYaml(self, sort=False) -> str:
+    def asYaml(self, sortKeys=False) -> str:
         """
         Returns this dict as yaml str, with comments, defaults, etc.
         """
         return _asYaml(self, doc=self._docs, validator=self._validator,
-                       default=self.default, sort=sort)
+                       default=self.default, sortKeys=sortKeys)
 
 
 class ConfigDict(CheckedDict):
@@ -649,9 +679,13 @@ class ConfigDict(CheckedDict):
             should return **None** to allow modification, **any value** to modify the value, or
             **raise ValueError** to stop the transaction
 
+        sortKeys: if True, keys are sorted whenever the dict is saved/edited. 
 
-    Example::
 
+    Example
+    =======
+
+    .. code::
 
         config = ConfigDict("myproj.subproj")
         config.addKey("keyA", 10, doc="documentaion of keyA")
@@ -699,7 +733,7 @@ class ConfigDict(CheckedDict):
                  persistent=True,
                  load=True,
                  fmt='yaml',
-                 sortKeys=True) -> None:
+                 sortKeys=False) -> None:
 
         if name:
             name = _normalizeName(name)
@@ -800,7 +834,7 @@ class ConfigDict(CheckedDict):
             logger.error(
                     f"Reset the dict to a default by removing the file '{self.getPath()}'"
             )
-            raise ValueError("dict is invalid")
+            raise ValueError(f"dict is invalid: {errormsg}")
         self._persistent, persistent = False, self._persistent
         super().update(kws)
         self._persistent = persistent
@@ -888,7 +922,7 @@ class ConfigDict(CheckedDict):
             fmt = self.fmt
         else:
             fmt = os.path.splitext(path)[1][1:]
-            assert fmt in {'json', 'yaml'}
+            assert fmt in {'json', 'yaml', 'csv'}
 
         logger.debug(f"Saving config to {path}")
         if fmt is None:
@@ -899,11 +933,49 @@ class ConfigDict(CheckedDict):
         elif fmt == 'yaml':
             yamlstr = self.asYaml(sortKeys=self.sortKeys)
             open(path, "w").write(yamlstr)
+        elif fmt == 'csv':
+            csvstr = self.asCsv()
+            open(path, "w").write(csvstr)
 
     def dump(self):
         """ Dump this config to stdout """
         print(str(self))
 
+    def _asRows(self):
+        rows = []
+        for key, value in self.items():
+            infostr = self._infoStr(key)
+            doc = self.getDoc(key)
+            rows.append((key, str(value), infostr, doc if doc else ""))
+        return rows
+
+    def asCsv(self) -> str:
+        """
+        Returns this dict as a csv str, with columns: key, value, spec, doc
+        """
+        rows = [("# key", "value", "spec", "doc")]
+        rows.extend(self._asRows())
+        from io import StringIO
+        import csv
+        s = StringIO()
+        writer = csv.writer(s)
+        writer.writerows(rows)
+        return s.getvalue()
+        
+    def _infoStr(self, k: str) -> str:
+        info = []
+        choices = self.getChoices(k)
+        if choices:
+            choicestr = "choices: {" + " ".join(str(ch) for ch in choices) + "}"
+            info.append(choicestr)
+        keyrange = self.getRange(k)
+        if keyrange:
+            low, high = keyrange
+            info.append(f"between {low} - {high}")
+        typestr = self.getTypestr(k)
+        info.append(typestr)
+        return" | ".join(info) if info else ""
+        
     def __str__(self) -> str:
         import tabulate
         header = f"Config: {self._name}\n"
@@ -911,23 +983,10 @@ class ConfigDict(CheckedDict):
         keys = sorted(self.keys())
         for k in keys:
             v = self[k]
-            info = []
             lines = []
-            choices = self.getChoices(k)
-            if choices:
-                choicestr = ", ".join(str(ch) for ch in choices)
-                if len(choicestr)>self._helpwidth:
-                    choiceslines = textwrap.wrap(choicestr, self._helpwidth)
-                    lines.extend(choiceslines)
-                else:
-                    info.append(choicestr)
-            keyrange = self.getRange(k)
-            if keyrange:
-                info.append(f"between {keyrange}")
-            typestr = self.getTypestr(k)
-            info.append(typestr)
+            infostr = self._infoStr(k)
             valuestr = str(v)
-            rows.append((k, valuestr, " | ".join(info)))
+            rows.append((k, valuestr, infostr))
             doc = self.getDoc(k)
             if doc:
                 doclines = textwrap.wrap(doc, self._helpwidth)
@@ -948,12 +1007,14 @@ class ConfigDict(CheckedDict):
         used is *yaml*, because it allows to embed comments. This is independent
         of the format used for persistence. The application used is the user's
         default application for the .yaml format and can be configured at the
-        os level. In macos we use `open`, in linux `xdg-open` and in windows
-        `start`, which all respond to the user's own configuration regarding
+        os level. In macos we use ``open``, in linux ``xdg-open`` and in windows
+        ``start``, which all respond to the user's own configuration regarding
         default applications.
 
-        NB: a temporary file is created for editing. The persisted file is only
-        modified if the editing is accepted.
+        .. note::
+
+            A temporary file is created for editing. The persisted file is only
+            modified if the editing is accepted.
 
         Args:
             waitOnModified: if True, the transaction is accepted whenever the
@@ -980,53 +1041,51 @@ class ConfigDict(CheckedDict):
         if configpath is None:
             configpath = self.getPath()
         if not os.path.exists(configpath):
-            if self.default is None:
-                logger.error(
-                        "No written config found, but default was not set")
-                raise FileNotFoundError(f"{configpath} not found")
             logger.debug("Using default config")
-            confdict = self.default
-        else:
-            logger.debug(f"Reading config from disk: {configpath}")
-            if self.default is None:
-                raise ValueError("Default config not set")
+            super().update(self.default)
+            return
+        logger.debug(f"Reading config from disk: {configpath}")
+        if self.default is None:
+            raise ValueError("Default config not set")
 
-            fmt = os.path.splitext(configpath)[1]
-            if fmt == ".json":
-                try:
-                    confdict = json.load(open(configpath))
-                except json.JSONDecodeError:
-                    error = sys.exc_info()[0]
-                    logger.error(f"Could not read config {configpath}: {error}")
-                    logger.debug("Using default as fallback")
-                    confdict = self.default
-            elif fmt == ".yaml":
-                try:
-                    with open(configpath) as f:
-                        confdict = yaml.load(f, Loader=yaml.SafeLoader)
-                except:
-                    logger.error(f"Could not read config {configpath}")
-                    confdict = self.default
-            else:
-                raise ValueError(f"format {fmt} unknown, supported formats: json, yaml")
+        fmt = os.path.splitext(configpath)[1]
+        if fmt == ".json":
+            try:
+                confdict = json.load(open(configpath))
+            except json.JSONDecodeError:
+                error = sys.exc_info()[0]
+                logger.error(f"Could not read config {configpath}: {error}")
+                logger.debug("Using default as fallback")
+                confdict = self.default
+        elif fmt == ".yaml":
+            try:
+                with open(configpath) as f:
+                    confdict = yaml.load(f, Loader=yaml.SafeLoader)
+            except:
+                logger.error(f"Could not read config {configpath}")
+                confdict = self.default
+        else:
+            raise ValueError(f"format {fmt} unknown, supported formats: json, yaml")
 
         # only keys in default should be accepted, but keys in the read
         # config should be discarded with a warning
         keysOnlyInRead = confdict.keys()-self.default.keys()
         if keysOnlyInRead:
             logger.warning(f"ConfigDict {self._name}, saved at {configpath}")
-            logger.warning(
-                    "There are keys defined in the saved"
-                    " config which are not present in the default config. They will"
-                    " be skipped:")
+            logger.warning("There are keys defined in the saved config which are not" 
+                           " present in the default config. They will be skipped:")
             logger.warning(f"   {keysOnlyInRead}")
 
         # merge strategy:
         # * if a key is shared between default and read dict, read dict has priority
         # * if a key is present only in default, it is added
-        confdict = _mergeDicts(confdict, self.default)
-        self.checkDict(confdict)
-        super().update(confdict)
+        
+        super().update(self.default)
+        errormsg = self.checkDict(confdict)
+        if errormsg:
+            logger.error(f"Could not load saved dict: {errormsg}. Using default")
+        else:
+            super().update(confdict)
         self._loaded = True
 
 
